package Text::Markdown::TreeBuilder;
use strict;
use warnings;
use base qw(HTML::Element);
use HTML::TreeBuilder;
use Text::Markdown::Parser;
use Data::Dumper;
use Encode;

=encoding utf8

=head1 NAME

Text::Markdown::TreeBuilder -- parse Markdown and generate (DOM like) Tree

=head1 SYNOPSIS

 use strict;
 use warnings;
 use Text::Markdown::TreeBuilder;

 my $tb=Text::Markdown::TreeBuilder->new(utf8=>1);
 $tb->parse(<<'TXT');
 AT&T has an ampersand in their name.
 =====================================

 AT&amp;T is another way to write it.
 ------------------------------------

 This & that.
 TXT
 print $tb->as_HTML();

=head1 DESCRIPTION

Parser for markdown formated Documents

=head1 AUTHOR

ToPeG

=head1 Overview

the Markdown types are converted to:

 markdown types:    perlclass:
  list_plain     -> HTML::Element "ul"
  list_count     -> HTML::Element "ol"
  list_entry     -> HTML::Element "li"
  h1..h6         -> HTML::Element "h1..h6"
  hr             -> HTML::Element "hr"
  link/linkref   -> HTML::Element "a"
  image/imageref -> HTML::Element "img"
  page           -> HTML::Element "p"
  code           -> HTML::Element "pre" => HTML::Element "code"
  inlinecode     -> HTML::Element "code"
  html           -> HTML::Element ( HTML::TreeBuilder ->look_down('body')->content->[0] )
  inlinehtml     -> HTML::Element ( HTML::TreeBuilder ->look_down('body')->content->[0] )
  quote          -> HTML::Element "blockquote"
  strong         -> HTML::Element "strong"
  em             -> HTML::Element "em"
  break          -> HTML::Element "br"
  comment        -> IGNORED
  link_def       -> IGNORED (only used in Text::Markdown::Parser to Determin imageref and linkref)

all delivered Methods from HTML::Element are usable

=cut

# CHANGELOG:
#
# 0.1 initial Release

our $VERSION='0.1';

=head1 METHODS

=over 4

=item new(%options)

create a new Parser Object

returns a Text::Markdown::TreeBuilder Object

=over 2

 type_attr => { type=>{...} }

 default:

 type_attr => {
    inlinecode => {style => "border:0.1em solid gray;" },
    code       => {style => "border:0.1em solid gray; margin-left:2em;" },
    quote      => {style => "border-left:0.3em solid blue; padding-left:0.3em; margin-left:0.6em;" },
  }

set attributes for HTML-Tags

 parser

set your own parser

 parser_tabwidth => <int> (default 4)

set the amount of spaces for one tab

 parser_grep_exessive => 1/0 (default 0)

if this option is true, linlinehtml will gab as mutch as possible text

=back

=cut

sub new
{
  my $class = shift;
  $class = ref($class) || $class;

  # Initialize HTML::Element part
  my $self = HTML::Element->new('html');

  $self->{'_store_comments'} = 0;
  $self->{'_init_parser'}    = 1;

  $self->{_type_attr}->{inlinecode}={style=>"border:0.1em solid gray;"};
  $self->{_type_attr}->{code}      ={style=>"border:0.1em solid gray; margin-left:2em;"};
  $self->{_type_attr}->{quote}     ={style=>"border-left:0.3em solid blue; padding-left:0.3em; margin-left:0.6em;"};

  my %p_opts=();

  # Parse attributes passed in as arguments
  if (@_)
  {
    pop(@_) if(@_%2);
    my %attr = @_;
    while ( my($k,$v)=each(%attr) )
    {
      if($k=~s/^parser_//)
      { $p_opts{$k}=$v;  }
      else
      { $self->{"_$k"} = $v; }
    }
  }

  $self->{'_parser'}=Text::Markdown::Parser->new(%p_opts) unless($self->{'_parser'});

  return bless($self,$class);
}

sub _elem
{
  my ( $self, $elem, $val ) = @_;
  if(exists($self->{$elem}))
  {
    my $old = $self->{$elem};
    $self->{$elem} = $val if defined $val;
    return $old;
  }
  return undef;
}

# accessors....
#sub store_comments  { shift->_elem( '_store_comments', @_ ); }

=item parser($parser)

set/get parser-opject

=cut

sub parser          { shift->_elem( '_parser', @_ ); }

=item utf8(1/0)

set/get UTF-8 conversion

=cut

sub utf8            { shift->_elem( '_utf8', @_ ); }

=item set_type_attributes($type,$param)

set the HTML-Attributes for a markdown type

see type_attr in "new"

=cut

sub set_type_attributes
{
  my $self=shift;
  my $type=shift;
  my $param=shift;
  return 0 unless($type && $param && ref($param) eq 'HASH');
  $self->{_type_attr}->{$type}=$param;
  return 1;
}

=item parse_file($file)

parse data from a File or Filehandle. See L<Text::Markdown::Parser>

=cut

sub parse_file
{
  my($self, $file) = @_;
  if (!ref($file) && ref(\$file) ne "GLOB")
  {
    open( $file , "<", $file) || return 0;
    binmode($file);
  }

  local $/=undef;
  return $self->parse(<$file>);
}

=item parse($text)

Parse Makdown formated Text. See L<Text::Markdown::Parser>

=cut

sub parse
{
  my $self=shift;
  my $txt=shift;

  $self->{_tree}=['body',{}];
  $self->{_elm_now}=[['root',$self->{_tree}]];

  if($self->{'_init_parser'})
  {
    $self->{'_init_parser'}=0;

    $self->{'_parser'}->add_handler('start',sub{
        my ($parser,$type,$url,$title,$alt,$id)=@_;

        my $element=['',{}];
        my $attr=$self->{_type_attr}->{$type};
        $element->[1]=$attr if($attr && ref($attr) eq 'HASH');

        if($type eq 'page')
        { $element->[0]='p'; }
        elsif($type=~/^h\d$/) # h1..h6
        { $element->[0]=$type; }
        elsif($type eq 'comment')
        {
          return 1 # no way to store comments jet....
          #return 1 unless($self->store_comments());
          #$element[0]='_comment';
        }
        elsif($type eq 'html' || $type eq 'inlinehtml')
        {
          $element->[0]='html';
          # setup special handler
          $parser->add_handler('text',sub{
              my ($type,$text)=@_;
              return 0 if($type ne 'html');
              my $htree=HTML::TreeBuilder->new();
              $htree->parse($text);
              my $html=$htree->look_down(_tag=>'body')->content->[0];
              $self->{_elm_now}->[0]->[1]=$html->clone();
              $htree->delete();
              return 0;
            },'type,text');
          #return 1;
        }
        elsif($type eq 'quote')
        { $element->[0]='blockquote'; }
        elsif($type eq 'code')
        {
          $type='code';
          unshift(@{$self->{_elm_now}},['code_pre',['pre',$element->[1]]]);
          $element->[0]='code';
          $element->[1]={};
        }
        elsif($type eq 'inlinecode')
        { $element->[0]='code'; }
        elsif($type eq 'hr')
        { $element->[0]='hr'; }
        elsif($type eq 'list_plain')
        { $element->[0]='ul'; }
        elsif($type eq 'list_count')
        { $element->[0]='ol'; }
        elsif($type eq 'list_entry')
        { $element->[0]='li'; }
        elsif($type eq 'break')
        { $element->[0]='br'; }
        elsif($type eq 'image' || $type eq 'imageref')
        {
          $element->[0]='img';
          $element->[1]->{src}=$url;
          $element->[1]->{alt}=$alt if($alt);
          $element->[1]->{title}=$title if($title);
        }
        elsif($type eq 'link' || $type eq 'linkref')
        {
          $element->[0]='a';
          $element->[1]->{href}=$url;
          $element->[1]->{title}=$title if($title);
        }
        elsif($type eq 'linkdef')
        {
          # ignore link definitions...
          return 1;
        }
        elsif($type eq 'strong')
        { $element->[0]='strong'; }
        elsif($type eq 'em')
        { $element->[0]='em'; }
        else
        {
          #unknown... (custom???)
          # ignore
          return 1;
        }

        #print "START: ".Dumper([$type,$element]);

        unshift(@{$self->{_elm_now}},[$type,$element]);

        return 1;
      },'self,type,url,title,alt,id');

    $self->{'_parser'}->add_handler('text' ,sub{
        my ($type,$text)=@_;
        #print "TEXT: ".Dumper([$type,$text]);
        return 1 if($type eq 'html');
        if($self->{_elm_now}->[0]->[0] eq $type)
        {
          $text=decode("utf8",$text) if($self->{_utf8});
          push(@{$self->{_elm_now}->[0]->[1]},$text);
        }
        return 1;
      },'type,text');

    $self->{'_parser'}->add_handler('end'  ,sub{
        my ($type)=@_;
        while(@{$self->{_elm_now}}>1)
        {
          my $elm=shift(@{$self->{_elm_now}});
          #print "END: ". Dumper($elm);
          # don't store empty "p"-Tags
          push(@{$self->{_elm_now}->[0]->[1]},$elm->[1]) if($type ne 'page' || @{$elm->[1]}>2);
          last if($elm->[0] eq $type);
        }
        return 1;
      },'type');
  }
  $self->{'_parser'}->parse($txt);
  #print Dumper($self->{_tree});
  $self->delete_content();
  my $body=HTML::Element->new_from_lol($self->{_tree});
  my $htree=['head',{}];
  push(@$htree,['meta',{'http-equiv'=>"Content-Type", content=>"text/html; charset=UTF-8"}]) if($self->{_utf8});
  my $head=HTML::Element->new_from_lol($htree);
  #print Dumper($t);
  $self->push_content($head);
  $self->push_content($body);
}

1;

=back

=cut

=head1 BUGS

unkown

=head1 TODO

- optional warnings

- store/export comments

=head1 See also

L<Text::Markdown::Parser>

=head1 AUTHOR

Tobias Grönhagen <markdown@topeg.de>

=head1 VERSION

Version 0.01 1 January 2011

=head1 COPYRIGHT

Copyright (c) Tobias Grönhagen. All rights reserved.

=head1 LICENSE

This program is free software; you can redistribute it
and/or modify it under the same terms and conditions as
Perl itself.

This means that you can, at your option, redistribute it and/or modify it under
either the terms the GNU Public License (GPL) version 2 or later, or under the
Perl Artistic License.

See http://dev.perl.org/licenses/

=head1 DISCLAIMER

THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS
OR IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
PARTICULAR PURPOSE.

Use of this software in any way or in any form, source or binary,
is not allowed in any country which prohibits disclaimers of any
implied warranties of merchantability or fitness for a particular
purpose or any disclaimers of a similar nature.

IN NO EVENT SHALL I BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
SPECIAL, INCIDENTAL,  OR CONSEQUENTIAL DAMAGES ARISING OUT OF THE
USE OF THIS SOFTWARE AND ITS DOCUMENTATION (INCLUDING, BUT NOT
LIMITED TO, LOST PROFITS) EVEN IF I HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE

=cut
